from datetime import datetime as dt
import logging
import os

import wave
import pyaudio
from playsound import playsound

from src.holon.HolonicAgent import HolonicAgent
from src.holon import logger


class Speaker(HolonicAgent):
    def __init__(self, cfg):
        super().__init__(cfg)


    def _on_connect(self, client, userdata, flags, rc):
        client.subscribe("voice.wave")

        super()._on_connect(client, userdata, flags, rc)


    def _on_message(self, client, db, msg):
        if "voice.wave" == msg.topic:
            try:
                filepath = dt.now().strftime("tests/_output/wave-%m%d-%H%M-%S.wav")
                with open(filepath, "wb") as file:
                    file.write(msg.payload)
                playsound(filepath)
                os.remove(filepath)
            except Exception as ex:
                logger.exception(ex)



# import atexit
import inspect
from multiprocessing import Process
import os
import signal
import sys
import threading
import time 
from typing import final
import uuid

currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))
parentdir = os.path.dirname(currentdir)
sys.path.insert(0, parentdir) 

import logging

from abdi_config import AbdiConfig, LOGGER_NAME
from broker.notifier import BrokerNotifier
from broker.broker_maker import BrokerMaker
from core.Agent import Agent
from holon.Blackboard import Blackboard
from holon.HolonicDesire import HolonicDesire
from holon.HolonicIntention import HolonicIntention
from holon.payload_wrapper import PayloadWrapper
from holon.payload import Payload


logger = logging.getLogger(LOGGER_NAME)



class HolonicAgent(Agent, BrokerNotifier) :
    def __init__(self, config:AbdiConfig=None, b:Blackboard=None, d:HolonicDesire=None, i: HolonicIntention=None):
        b = b or Blackboard()
        d = d or HolonicDesire()
        i = i or HolonicIntention()
        super().__init__(b, d, i)
        
        self.agent_id = None
        self.config = config if config else AbdiConfig(options={})
        self.head_agents = []
        self.body_agents = []
        self.__run_interval_seconds = 1
        
        self.name = f'<{self.__class__.__name__}>'
        self._agent_proc = None        
        self._broker = None
        self._topic_handlers = {}
        self._payload_wrapper = None


    @final
    def start(self, head=False):
        self._agent_proc = Process(target=self._run, args=(self.config,))
        self._agent_proc.start()
        
        for a in self.head_agents:
            a.start()
        for a in self.body_agents:
            a.start()
        
        if head:
            try:
                self._agent_proc.join()
            except:
                logger.warning(f"{self.name} terminated.")

    
    def unpack(self, payload):
        if self._payload_wrapper.is_managed(payload):
            return self._payload_wrapper.unpack(payload)
        else:
            return payload


# =====================
#  Instance of Process 
# =====================


    def is_running(self):
        return not self._terminate_lock.is_set()


    def _run(self, config:AbdiConfig):
        self.config = config
        self._run_begin()
        self._running()
        self._run_end()
    

    def _run_begin(self):
        self.on_begining()

        def signal_handler(signal, frame):
            logger.warning(f"{self.name} Ctrl-C: {self.__class__.__name__}")
            self.terminate()
        signal.signal(signal.SIGINT, signal_handler)

        self.agent_id = str(uuid.uuid1()).replace("-", "")
        self._terminate_lock = threading.Event()
        self._payload_wrapper = PayloadWrapper(self.agent_id)
        
        logger.debug(f"create broker")
        if broker_type := self.config.get_broker_type():
            self._broker = BrokerMaker().create_broker(
                broker_type=broker_type, 
                notifier=self)
            self._broker.start(options=self.config.options)
        
        logger.debug(f"start interval_loop")
        def interval_loop():
            while not self._terminate_lock.is_set():
                self.on_interval()
                time.sleep(self.__run_interval_seconds)
        threading.Thread(target=interval_loop).start()
            
        self.on_began()
        
        
    def get_run_interval(self):
        return self.__run_interval_seconds
        
        
    def set_run_interval(self, seconds):
        self.__run_interval_seconds = seconds


    def on_begining(self):
        pass


    def on_began(self):
        pass


    def on_interval(self):
        pass


    def _running(self):
        self.on_running()


    def on_running(self):
        pass


    def _run_end(self):
        self.on_terminating()

        while not self._terminate_lock.is_set():
            self._terminate_lock.wait(1)
        self._broker.stop() 
        
        self.on_terminated()


    def on_terminating(self):
        pass


    def on_terminated(self):
        pass


    @final
    def publish(self, topic, payload=None):
        return self._broker.publish(topic, payload)
        
        
    @final
    def request(self, topic, payload):
        logger.debug(f"payload: {len(payload)}")
        wrapped = self._payload_wrapper.wrap_for_request(payload)
        # logger.debug(f"wrapped: {len(wrapped)}")
        self.publish(topic, wrapped)
        
    
    def _on_response(self, topic, payload):
        managed_payload = self._payload_wrapper.unpack(payload)
        logger.debug(f"receiver: {managed_payload['receiver']}, agent_id: {self.agent_id}")
        if managed_payload["receiver"] == self.agent_id:
            self._on_message(topic, managed_payload["content"])
        
    
    def on_request(self, topic, payload):
        raise NotImplementedError()
        
    
    def _on_request(self, topic, payload):
        managed_payload = self._payload_wrapper.unpack(payload)
        logger.debug(f"topic: {topic}, managed_payload: {managed_payload}")

        handler = self._topic_handlers[topic] if topic in self._topic_handlers else self.on_request
        logger.debug(f"handler: {handler}")
        response_topic, response_payload = handler(topic, managed_payload["content"])
        logger.debug(f"response_topic: {response_topic}")
        if not response_payload:
            logger.warning(f"response_payload is None or empty.")
        
        if response_topic:
            managed_payload = self._payload_wrapper.wrap_for_response(response_payload, managed_payload)
            # logger.debug(f"response_payload: {response_payload}, managed_payload: {managed_payload}")
            self.publish(response_topic, managed_payload)
        

    @final
    def subscribe(self, topic, data_type="str", topic_handler=None):
        if topic_handler:
            logger.debug(f"Add topic handler: {topic}")
            self._topic_handlers[topic] = topic_handler
        return self._broker.subscribe(topic, data_type)
        

    @final
    def terminate(self):
        logger.warn(f"{self.name}.")

        for a in self.head_agents:
            name = a.__class__.__name__
            self.publish(topic='terminate', payload=name)

        for a in self.body_agents:
            name = a.__class__.__name__
            self.publish(topic='terminate', payload=name)

        self._terminate_lock.set()



# ==================================
#  Implementation of BrokerNotifier 
# ==================================

    
    def _on_connect(self):
        logger.info(f"{self.name} Broker is connected.")
        
        self.subscribe("echo")
        self.subscribe("terminate")
        self.on_connected()
            
            
    def on_connected(self):
        pass


    def _on_message(self, topic:str, payload):
        logger.debug(f"payload length: {len(payload)}")
        if payload and self._payload_wrapper.is_request(payload):
            logger.debug("message is_request")
            self._on_request(topic, payload)
        elif payload and self._payload_wrapper.is_response(payload):
            logger.debug(f"message is_response")
            self._on_response(topic, payload)
        else:
            logger.debug(f"unmanaged payload")
            if topic in self._topic_handlers:
                self._topic_handlers[topic](topic, payload)
            else:
                self.on_message(topic, payload)


    def on_message(self, topic:str, payload:Payload):
        pass



# ==================================
#  Others operation 
# ==================================


    def _convert_to_text(self, payload) -> str:
        if payload:
            data = payload.decode('utf-8', 'ignore')
        else:
            data = None
        
        return data
    
    
    def decode(self, payload):
        try:
            data = payload.decode('utf-8', 'ignore')
        except Exception as ex:
            logger.error(f"Type: {type(ex)}")
            logger.exception(ex)
            data = ""
            
        return data.strip()
        


import os, sys
sys.path.append(os.path.dirname(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__))))))

from datetime import datetime as dt
import threading

from src.holon import logger
import numpy as np
import pyaudio
import wave

from src.holon.HolonicAgent import HolonicAgent

# 配置音频录制参数
CHUNK = 2048
FORMAT = pyaudio.paInt16
CHANNELS = 1
RATE = 44100
RECORD_SECONDS = 5
MAX_RECORD_SECONDS = 10 * 60
SILENCE_THRESHOLD = (RATE // CHUNK) * 0.28

class Microphone(HolonicAgent):
    def __init__(self, cfg=None):
        super().__init__(cfg)


    def __compute_frames_mean(frames):
        def to_shorts(bytes):
            data = np.frombuffer(bytes, dtype=np.int16)
            return [x if x <= 32767 else 32767 - x for x in data]
        
        if not frames:
            return 0
        
        data = [x for x in to_shorts(frames) if x >= 0]
        audio_mean = 0 if len(data) == 0 else sum([int(x) for x in data]) // len(data)
        return audio_mean


    # def __is_silence(sound_raw):
    #     audio_mean = Microphone.__compute_frames_mean(sound_raw)
    #     #logging.debug(f'audio_mean: {audio_mean}')

    #     return (audio_mean < 200, audio_mean)
    

    def __wait_voice(self, audio_stream):
        first_frames = []
        logger.debug("for 60 second...")
        for _ in range(0, int(RATE / CHUNK * 60)):
            if not self.is_running():
                break
            # print(".", end="")
            try:
                sound_raw = audio_stream.read(CHUNK)
            except Exception as ex:
                logger.error("Read audio stream error!\n%s", str(ex))
                break

            if not Microphone.__compute_frames_mean(sound_raw) < 200:
                first_frames.append(sound_raw)  # 發現聲音
                if len(first_frames) > 2:
                    break                       # 確定開始發聲
            elif len(first_frames):
                    first_frames.clear()

        return first_frames if len(first_frames) else None
    

    def __record_to_silence(self, audio_stream):
        frames = []
        silence_count = 0
        total_mean = 0

        logger.debug("...")
        for i in range(0, int(RATE / CHUNK * MAX_RECORD_SECONDS)):
            if not self.is_running():
                break
            try:
                sound_raw = audio_stream.read(CHUNK)
            except Exception as ex:
                logger.error("Read audio stream error!\n%s", str(ex))
                break
            frames.append(sound_raw)

            mean = Microphone.__compute_frames_mean(sound_raw)
            total_mean += mean
            if mean < 200:
                silence_count += 1
                print('.', end='', flush=True)
            else:
                silence_count = 0
                print('^', end='', flush=True)
                # print(f'{mean}', end='', flush=True)
            if silence_count > SILENCE_THRESHOLD*1:
                print()
                logger.debug(f"silence_count:{silence_count}, frames: {len(frames)}")
                break

        frames_mean = total_mean // len(frames)
        logger.debug(f'frames mean: {frames_mean}')
        return frames, frames_mean


    def _record2(self):
        audio = pyaudio.PyAudio()
        audio_stream = audio.open(format=FORMAT, channels=CHANNELS,
                            rate=RATE, input=True,
                            frames_per_buffer=CHUNK)
        
        frames = self.__wait_voice(audio_stream)
        frames_mean = 0
        if frames:
            other_frames, frames_mean = self.__record_to_silence(audio_stream)
            frames.extend(other_frames)
        # logging.debug(f'Average frames: {Microphone.__compute_frames_mean([byte for sublist in frames for byte in sublist])}')
        # logging.debug(f'Frames: {frames}')

        # Stop recording
        audio_stream.stop_stream()
        audio_stream.close()
        audio.terminate()

        wave_path = None
        if frames and len(frames) >= SILENCE_THRESHOLD//2 and frames_mean >= 500:
            
            def write_wave_file(wave_path, wave_data):
                logger.info(f"Write to file: {wave_path}...")
                wf = wave.open(wave_path, 'wb')
                wf.setnchannels(CHANNELS)
                wf.setsampwidth(audio.get_sample_size(FORMAT))
                wf.setframerate(RATE)
                wf.writeframes(b''.join(frames))
                wf.close()
                self.publish("microphone.wave_path", wave_path)                
                # test
                #playsound(wave_path)
                #os.remove(wave_path)

            wave_path = dt.now().strftime("tests/_output/record-%m%d-%H%M-%S.wav")
            threading.Thread(target=write_wave_file, args=(wave_path, b''.join(frames),)).start()
            # self.publish("microphone.wave_path", wave_path)
            # write_wave_file(wave_path, b''.join(frames))

        return wave_path


    def _running(self):
        while self.is_running():
            try:
                # filepath = self._record()
                filepath = self._record2()
                # if filepath:
                    # with open(filepath, "rb") as file:
                    #     file_content = file.read()
                    # self.publish("hearing.microphone.voice", file_content)
                    # os.remove(filepath)
                    # logging.debug(f'Publish: microphone.wave_path: {filepath}')
                    # self.publish("microphone.wave_path", filepath)
            except Exception as ex:
                logger.exception(ex)


    def _record(self):
        # 初始化Pyaudio
        audio = pyaudio.PyAudio()

        # 开始录制音频
        stream = audio.open(format=FORMAT, channels=CHANNELS,
                            rate=RATE, input=True,
                            frames_per_buffer=CHUNK)

        frames = []
        logger.info("Start recording")
        for i in range(0, int(RATE / CHUNK * RECORD_SECONDS)):
            if (self.is_running()):
                # print(".", end="")
                data = stream.read(CHUNK)
                frames.append(data)

        # 停止录制音频
        stream.stop_stream()
        stream.close()
        audio.terminate()

        filepath = dt.now().strftime("tests/_output/record-%m%d-%H%M-%S.wav")
        def write_wave_file(filepath, wave_data):
            logger.debug("Write to file")
            wf = wave.open(filepath, 'wb')
            wf.setnchannels(CHANNELS)
            wf.setsampwidth(audio.get_sample_size(FORMAT))
            wf.setframerate(RATE)
            wf.writeframes(wave_data)
            wf.close()
        # threading.Thread(target=write_wave_file, args=(filepath, b''.join(frames),)).start()
        write_wave_file(filepath, b''.join(frames))

        return filepath


if __name__ == '__main__':
    logger.info('***** Microphone start *****')
    a = Microphone()
    a.start()


# if __name__ == '__main__':
#     logging.info('***** Microphone Test *****')

#     a = Microphone()

#     audio = pyaudio.PyAudio()
#     audio_stream = audio.open(format=FORMAT, channels=CHANNELS,
#                         rate=RATE, input=True,
#                         frames_per_buffer=CHUNK)
    
#     first_frames = a._wait_voice(audio_stream)
#     print(f"first_frames: {first_frames}")

#     # 停止录制音频
#     audio_stream.stop_stream()
#     audio_stream.close()
#     audio.terminate()



from paho.mqtt.client import Client

from broker.message_broker import MessageBroker
from broker.notifier import BrokerNotifier
import logging
from abdi_config import LOGGER_NAME


logger = logging.getLogger(LOGGER_NAME)


class MqttBroker(MessageBroker):
    def __init__(self, notifier:BrokerNotifier):
        self._client = Client()
        self.host = ""
        self.port = 0
        self.keepalive = 0
        
        super().__init__(notifier=notifier)


    def _on_connect(self, client:Client, userdata, flags, rc):
        logger.info(f"MQTT broker connected. url: {self.host}, port: {self.port}, keepalive: {self.keepalive}")
        # logger.debug(f"Client: {client}\nuserdata:{userdata}\nflags: {flags}\nrc: {rc}")
        self._notifier._on_connect()


    def _on_message(self, client:Client, db, message):
        try:
            self._notifier._on_message(message.topic, message.payload)
        except Exception as ex:
            logger.exception(ex)



    ###################################
    # Implementation of MessageBroker #
    ###################################
    
    
    def start(self, options:dict):
        logger.info(f"MQTT broker is starting...")
        
        self._client.on_connect = self._on_connect
        self._client.on_message = self._on_message
        if username := options.get("username"):
            self._client.username_pw_set(username, options.get("password"))
        
        self.host = options.get("host")
        self.port = options.get("port")
        self.keepalive = options.get("keepalive")
        # logger.debug(f"host:{host} port:{port} keep:{keepalive}")
        self._client.connect(self.host, self.port, self.keepalive)
        
        self._client.loop_start()


    def stop(self):
        logger.info(f"MQTT broker is stopping...")
        
        self._client.disconnect()
        self._client.loop_stop()


    def publish(self, topic:str, payload):
        return self._client.publish(topic=topic, payload=payload)
        
    
    def subscribe(self, topic:str, data_type):
        return self._client.subscribe(topic=topic)



import os
from pathlib import Path

import logging
from logging.handlers import TimedRotatingFileHandler


logger = logging.getLogger("ABDI")
__log_init = False


def init_logging(log_dir, log_level=logging.DEBUG):
    formatter = logging.Formatter(
        '%(levelname)1.1s %(asctime)s %(module)15s:%(lineno)03d %(funcName)15s) %(message)s',
        datefmt='%H:%M:%S')
    
    Path(log_dir).mkdir(parents=True, exist_ok=True)
    log_path = os.path.join(log_dir, "abdi.log")
    file_handler = TimedRotatingFileHandler(log_path, when="d")
    file_handler.setLevel(log_level)
    file_handler.setFormatter(formatter)

    console_handler = logging.StreamHandler()
    console_handler.setLevel(log_level)
    console_handler.setFormatter(formatter)
    
    logger = logging.getLogger("ABDI")
    logger.addHandler(console_handler)
    logger.addHandler(file_handler)    
    logger.setLevel(log_level)

    return logger


def get_logger():
    global __log_init
    global logger
    
    if not __log_init:
        logger = init_logging(log_dir="./_log")
        __log_init = True
        
    return logger



from abc import ABC, abstractmethod
from core import Belief 
from core import Desire 
from core import Intention 

class Agent(ABC):
    def __init__(self, b:Belief, d:Desire, i:Intention):
        self.B = b
        self.D = d
        self.I = i

    @abstractmethod
    def start(self):
        pass



import json
import logging
import uuid

from abdi_config import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)
VERSION = "0001"
# WRAPPER_HEAD = "950f7f7ba7c111eea5c4ff9ca9F3fcfd"
# WRAPPER_HEAD = "950f7f7ba7c111ee"
WRAPPER_REQUEST_HEAD = "950f7f7ba7c111ee"
WRAPPER_RESPONSE_HEAD = "a5c4ff9ca9F3fcfd"
WRAPPER_REQUEST_HEAD_BYTES = WRAPPER_REQUEST_HEAD.encode("utf-8")
WRAPPER_RESPONSE_HEAD_BYTES = WRAPPER_RESPONSE_HEAD.encode("utf-8")
WRAPPER_HEAD_LENGTH = len(WRAPPER_REQUEST_HEAD)


class PayloadWrapper:
    def __init__(self, agent_id:str):
        self.agent_id = agent_id
        
        
    def is_request(self, payload:bytes):
        managed = False

        if payload and len(payload) >= WRAPPER_HEAD_LENGTH:
            head_bytes = payload[:WRAPPER_HEAD_LENGTH]
            managed = head_bytes == WRAPPER_REQUEST_HEAD

        return managed
        
        
    def is_request(self, payload:bytes):
        return self.__check_head(payload, WRAPPER_REQUEST_HEAD_BYTES)
        
        
    def is_response(self, payload:bytes):
        return self.__check_head(payload, WRAPPER_RESPONSE_HEAD_BYTES)
    
    
    def __check_head(self, payload:bytes, head_to_check):
        managed = False

        if payload:
            head_bytes = payload[:len(head_to_check)]
            managed = head_bytes == head_to_check

        return managed


    def unpack(self, payload) -> str:
        payload_text = payload.decode('utf-8')
        payload_json = json.loads(payload_text[WRAPPER_HEAD_LENGTH:])
        # logger.debug(f"payload_json: {payload_json}")
        if VERSION != payload_json["version"]:
            raise Exception("Invalid payload version.")
        return payload_json


    def wrap_for_response(self, payload, managed_request_payload:str) -> str:
        receiver_agent_id = managed_request_payload["sender"]
        resp_json = {
            "version": VERSION,
            "receiver": receiver_agent_id
        }
        resp_json["content"] = payload

        resp = f"{WRAPPER_RESPONSE_HEAD}{json.dumps(resp_json)}"
        return resp


    def wrap_for_request(self, payload) -> str:
        request_json = {
            "version": VERSION,
            "sender": self.agent_id
        }
        request_json["content"] = payload

        request_payload = f"{WRAPPER_REQUEST_HEAD}{json.dumps(request_json)}"
        return request_payload



import json
import logging
import pickle

from abdi_config import LOGGER_NAME

logger = logging.getLogger(LOGGER_NAME)
VERSION = "0001"
VERSION_BYTES = VERSION.encode('utf-8')
# WRAPPER_HEAD = "950f7f7ba7c111eea5c4ff9ca9F3fcfd"
# WRAPPER_HEAD = "950f7f7ba7c111ee"
WRAPPER_REQUEST_HEAD = "950f7f7ba7c111ee"
WRAPPER_RESPONSE_HEAD = "a5c4ff9ca9F3fcfd"
WRAPPER_REQUEST_HEAD_BYTES = WRAPPER_REQUEST_HEAD.encode("utf-8")
WRAPPER_RESPONSE_HEAD_BYTES = WRAPPER_RESPONSE_HEAD.encode("utf-8")


class PayloadWrapper:
    def __init__(self, agent_id:str):
        self.binary_wrapper = BinaryWrapper(self)
        self.text_wrapper = TextWrapper(self)
        self.agent_id = agent_id
        

    def get_payload_wrapper(self, payload):
        # logger.debug(f"get_payload_wrapper, payload: {payload}, type: {type(payload)}")
        if payload:
            if self.text_wrapper.is_acceptable(payload):
                return self.text_wrapper
            elif self.binary_wrapper.is_acceptable(payload):
                return self.binary_wrapper
            else:
                return None
        else:
            return payload
        
        
    def is_request(self, payload:bytes):
        return self.__check_head(payload, WRAPPER_REQUEST_HEAD_BYTES)
        
        
    def is_response(self, payload:bytes):
        return self.__check_head(payload, WRAPPER_RESPONSE_HEAD_BYTES)
    
    
    def __check_head(self, payload:bytes, head_to_check):
        managed = False

        if payload:
            head_bytes = payload[:len(head_to_check)]
            # logger.debug(f"head_bytes: {head_bytes}, head_to_check: {head_to_check}")
            managed = head_bytes == head_to_check

        return managed


    def unpack(self, payload) -> str:
        wrapper = self.get_payload_wrapper(payload)
        if wrapper:
            unpacked = wrapper.unpack(payload)
        else:
            raise Exception("Not managed payload.")
            
        return unpacked


    def wrap_for_response(self, payload:str, managed_request_payload:str) -> str:
        wrapper = self.get_payload_wrapper(payload)
        if wrapper:
            payload_resp = wrapper.wrap_for_response(payload, managed_request_payload)
        else:
            raise Exception(f"Unsupported payload type: {type(payload)}")

        return payload_resp


    def wrap_for_request(self, payload) -> str:
        wrapper = self.get_payload_wrapper(payload)
        if wrapper:
            payload_resp = wrapper.wrap_for_request(payload)
        else:
            raise Exception("Unsupported payload type.")

        return payload_resp



class BinaryWrapper:
    def __init__(self, payload_wrapper:PayloadWrapper):
        self.payload_wrapper = payload_wrapper
        
        
    def is_acceptable(self, payload):
        return isinstance(payload, bytes) or isinstance(payload, bytearray)

        
    def unpack(self, payload:bytes) -> bytes:
        payload_body = payload[len(WRAPPER_REQUEST_HEAD_BYTES):]
        payload_json = pickle.loads(payload_body)

        if VERSION != payload_json["version"]:
            raise Exception("Invalid payload version.")

        return payload_json


    def wrap_for_request(self, payload:bytes) -> bytes:
        request_json = {
            "version": VERSION,
            "sender": self.payload_wrapper.agent_id
        }
        request_json["content"] = payload
        
        request_payload = WRAPPER_REQUEST_HEAD_BYTES + pickle.dumps(request_json)
        return request_payload
     
        
    def wrap_for_response(self, payload:bytes, managed_request_payload:bytes) -> bytes:
        receiver_agent_id = managed_request_payload["sender"]
        request_json = {
            "version": VERSION,
            "receiver": receiver_agent_id
        }
        request_json["content"] = payload
        
        request_payload = WRAPPER_RESPONSE_HEAD_BYTES + pickle.dumps(request_json)
        return request_payload
        

    # def wrap_for_request(self, payload:bytes) -> bytes:
    #     head_items = [WRAPPER_REQUEST_HEAD_BYTES, 
    #                    VERSION_BYTES, 
    #                    self.payload_wrapper.agent_id.encode('utf-8')]
    #     head_bytes = b''.join(item for item in head_items)
        
    #     request_payload = head_bytes + payload
    #     return request_payload
            
class TextWrapper:
    def __init__(self, payload_wrapper:PayloadWrapper):
        self.payload_wrapper = payload_wrapper
        
        
    def is_acceptable(self, payload):
        if isinstance(payload, str):
            acceptable = True
        else:
            try:
                _ = payload.decode('utf-8')
                acceptable = True
            except UnicodeDecodeError:
                acceptable = False
            
        return acceptable

        
    def unpack(self, payload:str) -> str:
        payload_text = payload.decode('utf-8')
        payload_json = json.loads(payload_text[len(WRAPPER_REQUEST_HEAD):])
        # logger.debug(f"payload_json: {payload_json}")
        if VERSION != payload_json["version"]:
            raise Exception("Invalid payload version.")
        return payload_json


    def wrap_for_request(self, payload:str) -> str:
        request_json = {
            "version": VERSION,
            "sender": self.payload_wrapper.agent_id
        }
        request_json["content"] = payload

        request_payload = f"{WRAPPER_REQUEST_HEAD}{json.dumps(request_json)}"
        return request_payload
                
        
    def wrap_for_response(self, payload:str, managed_request_payload:str) -> str:
        receiver_agent_id = managed_request_payload["sender"]
        resp_json = {
            "version": VERSION,
            "receiver": receiver_agent_id
        }
        resp_json["content"] = payload

        resp = f"{WRAPPER_RESPONSE_HEAD}{json.dumps(resp_json)}"
        return resp



from holon.HolonicAgent import HolonicAgent
from visual.Camera import Camera
from visual.ImagePreprocessing import ImagePreprocessing

class Visual(HolonicAgent):
    def __init__(self, cfg):
        super().__init__(cfg)
        self.head_agents.append(Camera(cfg))
        self.body_agents.append(ImagePreprocessing(cfg))
